/* clock -- For each multiple-of-three, isolated coding indel between
*  Human and (Opossum + Elephant), and each other placental mammal, report
*  if the other mammal agrees with Human (h), agrees with Elephant (e), is
*  missing data at that position (m), or aligns but agrees with neither (n).
* 
*  command: clock chr.maf chr-exons chr-name [W=??] [debug]
* 
*  W (default 40) is the "window radius" where we require no indels on either
*  side of the indel of interest.
*/

#include "lib.h"
#include "maf.h"

#define W_DEFAULT 40	// no other gaps within this radius

struct species {
	char *name, *latin;
} S[] =  {
	{"Human", "hg"},
	{"Chimp", "panTro"},
	{"Rhesus", "rheMac"},
	{"Galago", "otoGar"},
	{"TreeShrew", "tupBel"},
	{"Mouse", "mm"},
	{"Rat", "rn"},
	{"Guineapig", "cavPor"},
	{"Rabbit", "oryCun"},
	{"Dog", "canFam"},
	{"Cat", "felCat"},
	{"Horse", "equCab"},
	{"Cow", "bosTau"},
	{"Shrew", "sorAra"},
	{"Hedgehog", "eriEur"},
	{"Armadillo", "dasNov"},
	{"Tenrec", "echTel"},
	{"Elephant", "loxAfr"},
	{"Opossum", "monDom"},
#ifdef Never
	{"Platypus", "ornAna"},
	{"Chicken", "galGal"},
	{"Lizard", "anoCar"},
	{"Frog", "xenTro"},
	{"Fugu", "fr"},
	{"Tetraodon", "tetNig"},
	{"Stickleback", "gasAcu"},
	{"Medaka", "oryLat"},
	{"Zebrafish", "danRer"},
#endif
};

#define N_PLACENTALS 18

// find index of the common name of a species
int name2int(char *name) {
	int i, size = sizeof(S)/sizeof(S[0]);

	for (i = 0; i < size; ++i)
		if (same_string(S[i].name, name))
			return i;
	fatalf("name_to_int: %s not found", name);
}

// translate a common name to a latin name
char *name2latin(char *name) {
	return S[name2int(name)].latin;
}

// data structure for interval of radius W around a coding indel
struct candidate {
	int hb, he,	// human region boundaries
	    db, de;	// elephant region boundaries
} C[10000];

#define HUGE 1000000
char human[HUGE], elephant[HUGE];
struct mafFile *mf;
struct mafAli *A;
int dstart, N, debug, W = W_DEFAULT;

// find the maf component for a named species
struct mafComp *find_comp(struct mafAli *a, char *name) {
	struct mafComp *c;
	int len;
	char *species = name2latin(name);

	len = strlen(species);
	for (c = a->components;
	     c != NULL && strncmp(c->src, species, len);
	     c = c->next)
		;
	return c;
}

// tell if there is a gap between species1 and species2 for
// species1-positions x with b <= x < e
// 2 = missing data, 1 = data exists and has a gap, 0 = no gap
int gap(char *species1, char *species2, int b, int e) {
	struct mafAli *a;
	struct mafComp *s1, *s2;
	int i, pos, critical = 0;
	char x, y;

// I found this debugging hack to be quite useful, so it remains.   --Webb
// int dd = (same_string(species1, "Human") && same_string(species2, "TreeShrew"));

int dd = 0;

if (dd) printf("dd: %s %d %d vs %s: ", species1, b, e, species2);
	if (debug)
		printf("%s vs %s: ", species1, species2);
	for (a = A; a != NULL; a = a->next) {
		s1 = find_comp(a, species1);
		s2 = find_comp(a, species2);
		if (s1 == NULL) {
			if (critical && s2 != NULL)
				return 1;
			continue;
		}
		if (critical && s1->start >= e)
			break;
		if (s1->start + s1->size <= b)
			continue;
		if (s2 == NULL)
			return 2;
if (dd) {
printf("maf: %d %d\n", s1->start, s1->start + s1->size);
for (i = 0; i < a->textSize; ++i)
putchar(s1->text[i]);
putchar('\n');
for (i = 0; i < a->textSize; ++i)
putchar(s2->text[i]);
putchar('\n');
putchar('\n');
}

		for (pos = s1->start - 1, i = 0; i < a->textSize; ++i) {
			if (s1->text[i] != '-')
				++pos;
			if (pos >= b && pos < e) {
				critical = 1;
				x = s1->text[i];
				y = s2->text[i];
if (dd) printf("  %c%c\n", x, y);
				if ((x == '-' && y != '-') ||
				    (x != '-' && y == '-')) {
				    	if (debug)
						printf("gap\n");
if (dd) printf("gap\n");
					return 1;
				}
			}
		}
	}
	if (debug)
		printf(" no gap\n");
	return 0;
}

// categorize a species as like human (h), like elephant (e), like neither (n),
// or missing data (m)
char bin_species(int i, char *species) {
	int g;

	if ((g = gap("Human", species, C[i].hb, C[i].he)) == 0)
		return 'h';
	if (g == 2)
		return 'm';
	if ((g = gap("Elephant", species, C[i].db, C[i].de)) == 0)
		return 'e';
	if (g == 2)
		return 'm';
	return 'n';
}

// check a candidate window
void ck_cand(int i, char *chr, int B, int E) {
	int b, e, k;

	b = C[i].db;
	e = C[i].de;
	if (debug)
		printf("checking elephant %d-%d\n", b, e);
	if (gap("Elephant", "Opossum", b, e) != 0)
	    	return; // nothing
	for (k = 0; k < N_PLACENTALS; ++k)
		putchar(bin_species(i, S[k].name));
	printf("\t%s %d %d (%d bp, near %d)\n", chr, B, E, E-B, C[i].hb+W);
}
		
// scan  mafs covering a human exon; save non-dash columns for Human, Elephant
int hom_ele(int b, int e) {
	struct mafAli *a;
	struct mafComp *h, *d;
	int i, hpos, next_dpos;
	char *ht, *dt, ele_src[100];

	// find the maf containing human position b
	for (a = A ; a != NULL; a = a->next) {
		h = a->components;
		if ( h->start + h->size > b)
			break;
	}
	if (h->start > b)
		return 0;
	ht = h->text;

	d = find_comp(a, "Elephant");
	if (d == NULL)
		return 0;
	next_dpos = d->start + d->size;
	dt = d->text;
	strcpy(ele_src, d->src);
 
	// find alignment column containing human position b
	hpos = h->start - 1;
	dstart = d->start - 1;
	for (i = 0; i < a->textSize; ++i) {
		if (dt[i] != '-')
			++dstart;
		if (ht[i] != '-' && ++hpos == b)
			break;
	}
	if (i == a->textSize)
		fatal("ran off human text in hom_ele");
	// dstart is now the elephant position corresponding to human position b
//printf("dstart = %d\n", dstart);

	// now hpos = b; in what follow, hpos is the next human position
	// record non-dash columns until human position e or end of maf
	for (N = 0; i < a->textSize; ++i) {
		if (ht[i] != '-' || dt[i] != '-') {
			human[N] = ht[i];
			elephant[N] = dt[i];
			if (++N >= HUGE)
				fatal("huge exon");
		}
		if (ht[i] != '-' && ++hpos == e)
			break;
	}

	// if necessary, read more mafs to reach human position e
//print_align();
//printf("pos = %d, e = %d\n", pos, e);
	while (hpos < e) {
		// like for the first maf, minus scanning for position b,
		// plus a contiguity check for Human and Elephant
		if ((a = a->next) == NULL)
			return 0;
		h = a->components;
		ht = h->text;
		d = find_comp(a, "Elephant");
		if (d == NULL)
			return 0;
		if (d->start != next_dpos)
			return 0;
		next_dpos = d->start + d->size;
		dt = d->text;
		for (i = 0; i < a->textSize; ++i) {
			if (ht[i] != '-' || dt[i] != '-') {
				human[N] = ht[i];
				elephant[N] = dt[i];
				if (++N >= HUGE)
					fatal("huge exon");
			}
			if (ht[i] != '-' && ++hpos == e)
				break;
		}
//print_align();
//printf("pos = %d, e = %d\n", pos, e);
//exit(0);
	}

	return 1;
}

// find candidate windows
int get_candidates(int b, int e) {
	int i, j, k, hpos, dpos, n = 0, gap_beg, gap_end, hgap, good;

	if (hom_ele(b, e) == 0)	// exon does not align properly
		return 0;
	if (debug) {
		human[N] = elephant[N] = '\0';
		printf("human-elephant alignment (%d columns):\n  %s\n  %s\n", N, human, elephant);
	}
	hpos = b;
	dpos = dstart;
	// at column i of human/elephant (pairwise alignment) aligns human position
	// hpos with elephant position dpos
	// find gap_start > i
	for (i = 0; i < N; ) {
		// find next gap
		good = 1;
		for (j = i; j < N && human[j] != '-' && elephant[j] != '-'; ++j)
			;
		if (j == N)
			break;
		gap_beg = j;
		if (gap_beg < i+W)
			good = 0;
		if ((hgap = (human[j] == '-')) && elephant[j] == '-')
			fatal("column of dashes");

		// find gap end
		while (human[j] == '-')
			++j;
		while (elephant[j] == '-')
			++j;
		gap_end = j;
		if (debug)
			printf("  gap in columns %d-%d\n", gap_beg, gap_end);
		if ((gap_end - gap_beg)%3 != 0) {
			if (debug)
				printf("length not divisible by 3\n");
			good = 0;
		}
		for (k = gap_beg; k < gap_end; ++k)
			if ((hgap == 1 && elephant[k] == '-') ||
			    (hgap == 0 && human[k] == '-'))
				break;
		if (k < gap_end) {
			if (debug)
				printf("  fails k = %d < gap_end = %d\n",
				  k, gap_end);
			good = 0;
		}
		while (j < N && j < gap_end+W && human[j] != '-' &&
		    elephant[j] != '-')
			++j;
		if (j < gap_end+W) {
			if (debug)
				printf("fails j = %d < gap_end + W = %d + %d\n",
				  k, gap_end, W);
			good = 0;
		} else if (j == N) {
			if (debug)
				printf("  too near the end of the alignment\n");
			break;
		}
		for ( ; i < gap_end; ++i) {
			if (i == gap_beg && good) {
				C[n].hb = hpos - W;
				C[n].db = dpos - W;
			}
			if (human[i] != '-')
				++hpos;
			if (elephant[i] != '-')
				++dpos;
		}
		if (good) {
			if (debug)
				printf("gap is good\n");
			C[n].he = hpos + W;
			C[n].de = dpos + W;
			++n;
		}
	}

	return n;
}

// Keep a list of mafs that cover the interval [b,e], guaranteeing that the
// first component of each maf is human.
int get_maf_list(int b, int e, int fuzz) {
	struct mafAli *a, *end;
	struct mafComp *h;
	char *hum = name2latin("Human");
	int len;
	static int is_EOF = 0;

	if (is_EOF)
		return 0;
	len = strlen(hum);

	if (debug)
		printf("maf_list: b = %d, e = %d, fuzz = %d\n", b, e, fuzz);
	if (A != NULL && (h = A->components) != NULL &&
	    h->start > b)
	    	return 3;
	// flush useless mafs from the start of the list
	while (A != NULL) {
		h = A->components;
		if (h->start + h->size > b - fuzz)
		    	break;
		a = A;
		A = A->next;
		mafAliFree(&a);
	}
	if (A != NULL && h->start > b) {
		if (debug)
			printf("gap in mafs\n");
		return 2;
	}

	// find the end of the current list
	if ((end = A) != NULL)
		while (end->next != NULL)
			end = end->next;

	// add mafs until the interval, plus a little extra, is covered
	while ((a = mafNext(mf)) != NULL) {
		h = a->components;
		// guarantee that first component is human
		if (h == NULL || strncmp(h->src, hum, len))
			fatal("first maf component is not human");
		if (debug)
			printf(" maf %s %d %d: ",
			   h->src, h->start, h->start + h->size);
		if (h->start + h->size <= b - fuzz) {
			if (debug)
				printf("free\n");
			mafAliFree(&a);
		} else {
			if (debug)
				printf(" add to list\n");
			if (A == NULL)
				A = a;
			else
				end->next = a;
			end = a;
			if (h->start > e + fuzz)
				break;
		}
	}

	if (debug) {
		if (A != NULL) {
			h = A->components;
			printf("maf list from [%d-%d] ",
		  	h->start, h->start + h->size);
			h = end->components;
			printf("to [%d-%d]\n", h->start, h->start + h->size);
		} else
			printf("no maf list\n");
	}
	is_EOF = (a == NULL);

	return (A != NULL);
}

int main(int argc, char **argv) {
	FILE *fp;
	int b, e, old_e = 0, g, i, ncand;
	char buf[500], chr[100];

	if (argc > 4 && same_string(argv[argc-1], "debug")) {
		debug = 1;
		--argc;
	}
	if (argc > 4 && strncmp(argv[argc-1], "W=", 2) == 0) {
		W = atoi(argv[--argc]+2);
		fprintf(stderr, "W reset to %d\n", W);
	}
	if (argc != 4)
		fatalf("args: chr.maf chr-exons chr-name [W=??] [debug]");

	mf = mafOpen(argv[1], 0);
	fp = ckopen(argv[2], "r");
	while (fgets(buf, 500, fp)) { 
		if (sscanf(buf, "%s %d %d", chr, &b, &e) != 3)
			fatalf("ends: %s", buf);
		if (!same_string(chr, argv[3]))
			fatalf("%s != %s", chr, argv[3]);
		if (b < old_e)
			//fatalf("%d %d out of order on %s", old_e, b, argv[2]);
			continue;
		old_e = e;
		if (debug)
			printf("---- trying %s", buf);
		g = get_maf_list(b, e, 50);
		if (g == 0)
			break;
		if (g > 1) {
			if (g == 2)
				fprintf(stderr, "oddity at %s", buf);
			continue;
		}
		ncand = get_candidates(b, e);
		if (debug) {
			printf("%d candidates\n", ncand);
			for (i = 0; i < ncand; ++i)
				printf("human %d-%d, elephant %d-%d\n",
				 C[i].hb, C[i].he, C[i].db, C[i].de);
		}
		for (i = 0; i < ncand; ++i)
			ck_cand(i, argv[3], b, e);
	}
	return 0;
}
