// for a specified gene list (e.g., RefSeq) and chromosome, create a
// sorted, non-redundant file of exons 

#include "lib.h"

// upper bound on number of exons in any chromosome
#define MAX_EXONS 100000

struct exon {
	char *gene;
	int beg, end, strand, phase, phase2, which_exon, count_exons;
} Exon[MAX_EXONS];
int nexons;

// fill in the which_exon, count_exons and phase2 fields for each exon
void count_exons() {
	int i, j, next_i;
	char x;

	for (i = 0; i < nexons; i = next_i) {
		// find all exons for the current gene
		for (next_i = i+1;
		     next_i < nexons &&
		       same_string(Exon[i].gene, Exon[next_i].gene);
		     ++next_i)
			;
		x = Exon[i].strand;
		for (j = i; j < next_i; ++j) {
			Exon[j].which_exon =
			  (x == '+' ? j-i+1 : next_i-j);
			// number of exons for this gene
			Exon[j].count_exons = next_i-i;
			// phase of the next intron
			Exon[j].phase2 =
			  (j < next_i ? Exon[j+1].phase : 0);
		}
	}
}

// sort exons by start-position; in case of a tie, use the end-position
int compar(struct exon *a, struct exon *b) {
	int x = a->beg, y = b->beg, p = a->end, q = b->end;

	if (x < y)
		return -1;
	if (y < x)
		return 1;
	if (p < q)
		return -1;
	if (q < p)
		return 1;
	return 0;
}

int main(int argc, char **argv) {
	FILE *fp;
	char buf[50000], *s, *T, *t, x[100], y[100], z[100];
	int i, j, b, e, B, E, bad, tot_len, ngenes, nuniq;

	if (argc != 3)
		fatal("args: refseq-names chr-name");

	fp = ckopen(argv[1], "r");
	ngenes = nuniq = 0;
	while (fgets(buf, 50000, fp)) {
		if (sscanf(buf,
		    "%s %s %s %*s %*s %d %d", x, y, z, &B, &E) != 5 ||
		    !same_string(y, argv[2]))
			continue;
		++ngenes;
		for (s = buf, i = 0; s != NULL && i < 8;
		  s = strchr(s+1, '\t'), ++i)
			;
		if (s == NULL)
			fatalf("ran off end of %s", buf);
		if ((T = strchr(++s, '\t')) == NULL)
			fatalf("ends of exons: %s", buf);
		t = T+1;
		tot_len = 0;
		while (s != NULL && t != NULL && s < T) {
			b = MAX(atoi(s), B);
			e = MIN(atoi(t), E);
			if (b <= e) {
				Exon[nexons].gene = copy_string(x);
				Exon[nexons].beg = b;
				Exon[nexons].end = e;
				Exon[nexons].strand = z[0];
				Exon[nexons].phase = tot_len % 3;
				++nexons;
				tot_len += (e - b);
			}
			if ((s = strchr(s, ',')) != NULL)
				++s;
			if ((t = strchr(t, ',')) != NULL)
				++t;
		}
	}
	fclose(fp);
	count_exons();

	qsort(Exon, (size_t)nexons, sizeof(struct exon), (const void *)compar);

	for (nuniq = bad = i = 0; i < nexons; ) {
		++nuniq;
		for (j = i, ++i;
		   i < nexons && Exon[j].beg == Exon[i].beg &&
		   Exon[j].end == Exon[i].end;
		   ++i)
			if (Exon[j].strand != Exon[i].strand ||
			    Exon[j].phase != Exon[i].phase)
				if (Exon[j].phase >= 0) {
					fprintf(stderr,
					  "%d-%d: %s(%d) %s(%d)\n",
					  Exon[j].beg, Exon[j].end,
					  Exon[j].gene, Exon[j].phase,
					  Exon[i].gene, Exon[i].phase);
					Exon[j].phase = -1;
					Exon[i].phase = -1;
				}
		if (Exon[j].phase == -1)
			++bad;
		else
			printf("%s %d %d %c %s %d %d %d %d\n", argv[2],
			  Exon[j].beg, Exon[j].end, Exon[j].strand,
			  Exon[j].gene, Exon[j].which_exon, Exon[j].count_exons,
			  Exon[j].phase, Exon[j].phase2);
	}
	fprintf(stderr, "%d coding exons in %d genes; %d unique exons\n",
	  nexons, ngenes, nuniq);
	if (bad > 0)
		fprintf(stderr, "%d exons have inconsistent phase\n", bad);

	return 0;
}
